@@ -1,7 +1,7 @@
 /*
  * netutils.c - Network utilities
  *
- * Copyright (C) 2013 - 2016, Max Lv <max.c.lv@gmail.com>
+ * Copyright (C) 2013 - 2017, Max Lv <max.c.lv@gmail.com>
  *
  * This file is part of the shadowsocks-libev.
  *
@@ -23,21 +23,15 @@
 #include <math.h>
 
 #include <libcork/core.h>
-#include <udns.h>
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
 
-#ifdef __MINGW32__
-#include "win32.h"
-#define sleep(n) Sleep(1000 * (n))
-#else
-#include <sys/socket.h>
+#include <arpa/inet.h>
 #include <netdb.h>
 #include <netinet/in.h>
 #include <unistd.h>
-#endif
 
 #if defined(HAVE_SYS_IOCTL_H) && defined(HAVE_NET_IF_H) && defined(__linux__)
 #include <net/if.h>
@@ -54,17 +48,22 @@
 
 extern int verbose;
 
+static const char valid_label_bytes[] =
+    "-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz";
+
 #if defined(MODULE_LOCAL)
 extern int keep_resolving;
 #endif
 
-int set_reuseport(int socket)
+int
+set_reuseport(int socket)
 {
     int opt = 1;
     return setsockopt(socket, SOL_SOCKET, SO_REUSEPORT, &opt, sizeof(opt));
 }
 
-size_t get_sockaddr_len(struct sockaddr *addr)
+size_t
+get_sockaddr_len(struct sockaddr *addr)
 {
     if (addr->sa_family == AF_INET) {
         return sizeof(struct sockaddr_in);
@@ -75,32 +74,35 @@ size_t get_sockaddr_len(struct sockaddr *addr)
 }
 
 #ifdef SET_INTERFACE
-int setinterface(int socket_fd, const char *interface_name)
+int
+setinterface(int socket_fd, const char *interface_name)
 {
     struct ifreq interface;
-    memset(&interface, 0, sizeof(interface));
+    memset(&interface, 0, sizeof(struct ifreq));
     strncpy(interface.ifr_name, interface_name, IFNAMSIZ);
     int res = setsockopt(socket_fd, SOL_SOCKET, SO_BINDTODEVICE, &interface,
                          sizeof(struct ifreq));
     return res;
 }
+
 #endif
 
-int bind_to_address(int socket_fd, const char *host)
+int
+bind_to_address(int socket_fd, const char *host)
 {
     if (host != NULL) {
         struct cork_ip ip;
         struct sockaddr_storage storage;
-        memset(&storage, 0, sizeof(storage));
+        memset(&storage, 0, sizeof(struct sockaddr_storage));
         if (cork_ip_init(&ip, host) != -1) {
             if (ip.version == 4) {
                 struct sockaddr_in *addr = (struct sockaddr_in *)&storage;
-                dns_pton(AF_INET, host, &addr->sin_addr);
+                inet_pton(AF_INET, host, &addr->sin_addr);
                 addr->sin_family = AF_INET;
                 return bind(socket_fd, (struct sockaddr *)addr, sizeof(struct sockaddr_in));
             } else if (ip.version == 6) {
                 struct sockaddr_in6 *addr = (struct sockaddr_in6 *)&storage;
-                dns_pton(AF_INET6, host, &addr->sin6_addr);
+                inet_pton(AF_INET6, host, &addr->sin6_addr);
                 addr->sin6_family = AF_INET6;
                 return bind(socket_fd, (struct sockaddr *)addr, sizeof(struct sockaddr_in6));
             }
@@ -109,21 +111,24 @@ int bind_to_address(int socket_fd, const char *host)
     return -1;
 }
 
-ssize_t get_sockaddr(char *host, char *port, struct sockaddr_storage *storage, int block)
+ssize_t
+get_sockaddr(char *host, char *port,
+             struct sockaddr_storage *storage, int block,
+             int ipv6first)
 {
     struct cork_ip ip;
     if (cork_ip_init(&ip, host) != -1) {
         if (ip.version == 4) {
             struct sockaddr_in *addr = (struct sockaddr_in *)storage;
             addr->sin_family = AF_INET;
-            dns_pton(AF_INET, host, &(addr->sin_addr));
+            inet_pton(AF_INET, host, &(addr->sin_addr));
             if (port != NULL) {
                 addr->sin_port = htons(atoi(port));
             }
         } else if (ip.version == 6) {
             struct sockaddr_in6 *addr = (struct sockaddr_in6 *)storage;
             addr->sin6_family = AF_INET6;
-            dns_pton(AF_INET6, host, &(addr->sin6_addr));
+            inet_pton(AF_INET6, host, &(addr->sin6_addr));
             if (port != NULL) {
                 addr->sin6_port = htons(atoi(port));
             }
@@ -158,18 +163,24 @@ ssize_t get_sockaddr(char *host, char *port, struct sockaddr_storage *storage, i
             return -1;
         }
 
+        int prefer_af = ipv6first ? AF_INET6 : AF_INET;
         for (rp = result; rp != NULL; rp = rp->ai_next)
-            if (rp->ai_family == AF_INET) {
-                memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in));
+            if (rp->ai_family == prefer_af) {
+                if (rp->ai_family == AF_INET)
+                    memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in));
+                else if (rp->ai_family == AF_INET6)
+                    memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in6));
                 break;
             }
 
         if (rp == NULL) {
-            for (rp = result; rp != NULL; rp = rp->ai_next)
-                if (rp->ai_family == AF_INET6) {
+            for (rp = result; rp != NULL; rp = rp->ai_next) {
+                if (rp->ai_family == AF_INET)
+                    memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in));
+                else if (rp->ai_family == AF_INET6)
                     memcpy(storage, rp->ai_addr, sizeof(struct sockaddr_in6));
-                    break;
-                }
+                break;
+            }
         }
 
         if (rp == NULL) {
@@ -184,8 +195,9 @@ ssize_t get_sockaddr(char *host, char *port, struct sockaddr_storage *storage, i
     return -1;
 }
 
-int sockaddr_cmp(struct sockaddr_storage *addr1,
-                 struct sockaddr_storage *addr2, socklen_t len)
+int
+sockaddr_cmp(struct sockaddr_storage *addr1,
+             struct sockaddr_storage *addr2, socklen_t len)
 {
     struct sockaddr_in *p1_in   = (struct sockaddr_in *)addr1;
     struct sockaddr_in *p2_in   = (struct sockaddr_in *)addr2;
@@ -226,8 +238,9 @@ int sockaddr_cmp(struct sockaddr_storage *addr1,
     }
 }
 
-int sockaddr_cmp_addr(struct sockaddr_storage *addr1,
-                      struct sockaddr_storage *addr2, socklen_t len)
+int
+sockaddr_cmp_addr(struct sockaddr_storage *addr1,
+                  struct sockaddr_storage *addr2, socklen_t len)
 {
     struct sockaddr_in *p1_in   = (struct sockaddr_in *)addr1;
     struct sockaddr_in *p2_in   = (struct sockaddr_in *)addr2;
@@ -251,3 +264,40 @@ int sockaddr_cmp_addr(struct sockaddr_storage *addr1,
         return memcmp(addr1, addr2, len);
     }
 }
+
+int
+validate_hostname(const char *hostname, const int hostname_len)
+{
+    if (hostname == NULL)
+        return 0;
+
+    if (hostname_len < 1 || hostname_len > 255)
+        return 0;
+
+    if (hostname[0] == '.')
+        return 0;
+
+    const char *label = hostname;
+    while (label < hostname + hostname_len) {
+        size_t label_len = hostname_len - (label - hostname);
+        char *next_dot   = strchr(label, '.');
+        if (next_dot != NULL)
+            label_len = next_dot - label;
+
+        if (label + label_len > hostname + hostname_len)
+            return 0;
+
+        if (label_len > 63 || label_len < 1)
+            return 0;
+
+        if (label[0] == '-' || label[label_len - 1] == '-')
+            return 0;
+
+        if (strspn(label, valid_label_bytes) < label_len)
+            return 0;
+
+        label += label_len + 1;
+    }
+
+    return 1;
+}
